#!/usr/bin/env python


import random
import textwrap as tw
import os.path  as op

import numpy as np
import pandas as pd

import funpack.hierarchy as hierarchy

from . import (tempdir, gen_DataTable)


def test_getHierarchyFilePath():

    dt = gen_DataTable([np.random.randint(1, 10, 10)])

    dt.vartable.loc[1, 'DataCoding'] = 123

    exp = op.join(op.dirname(hierarchy.__file__),
                  'data', 'hierarchy', 'coding123.tsv')
    assert hierarchy.getHierarchyFilePath(dt, 1) == exp


def test_loadHierarchyFile():

    hier = tw.dedent("""
    coding	meaning	node_id	parent_id
    30	meaning 30	3	1
    40	meaning 40	4	1
    50	meaning 50	5	4
    20	meaning 20	2	0
    60	meaning 60	6	2
    10	meaning 10	1	0
    """).strip()

    with tempdir():

        with open('hier.txt', 'wt') as f:
            f.write(hier)

        h = hierarchy.loadHierarchyFile('hier.txt')

        print(h._Hierarchy__codings)

        assert h.parents(10) == []
        assert h.parents(20) == []
        assert h.parents(30) == [10]
        assert h.parents(40) == [10]
        assert h.parents(50) == [40, 10]
        assert h.parents(60) == [20]



def test_codeToNumeric():

    icd10code = hierarchy.HIERARCHY_DATA_NAMES['icd10']
    icd10hier = hierarchy.getHierarchyFilePath(coding=icd10code)
    icd10hier = hierarchy.loadHierarchyFile(icd10hier)
    codes     = list(random.sample(icd10hier.codings, 20)) + ['badcode']
    exp       = [icd10hier.index(c) + 1 for c in codes[:-1]] + [np.nan]

    conv = [hierarchy.codeToNumeric(c, coding=icd10code) for c in codes]

    exp  = np.array(exp)
    conv = np.array(conv)

    expna  = np.isnan(exp)
    convna = np.isnan(conv)

    assert np.all(     expna  ==       convna)
    assert np.all(exp[~expna] == conv[~convna])

    ncodes = [(code, ncode)
              for code, ncode
              in zip(codes, conv)
              if not pd.isna(ncode)]

    codes, ncodes = zip(*ncodes)
    codes = [c.lower() for c in codes]

    assert codes == [hierarchy.numericToCode(c, coding=icd10code).lower()
                     for c in ncodes]


def test_Hierarchy():
    with tempdir():
        data = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        a\ta desc\t5\t0
        b\tb desc\t1\t5
        c\tc desc\t3\t5
        d\td desc\t4\t3
        e\te desc\t2\t1
        """)

        with open('codings.tsv', 'wt') as f:
            f.write(data)

        h = hierarchy.loadHierarchyFile('codings.tsv')

        assert h.index('a') == 4
        assert h.index('b') == 0
        assert h.index('c') == 2
        assert h.index('d') == 3
        assert h.index('e') == 1
        assert h.coding(0)  == 'b'
        assert h.coding(1)  == 'e'
        assert h.coding(2)  == 'c'
        assert h.coding(3)  == 'd'
        assert h.coding(4)  == 'a'

        assert h.parents('a') == []
        assert h.parents('b') == ['a']
        assert h.parents('c') == ['a']
        assert h.parents('d') == ['c', 'a']
        assert h.parents('e') == ['b', 'a']

        assert h.description('a') == 'a desc'
        assert h.description('b') == 'b desc'
        assert h.description('c') == 'c desc'
        assert h.description('d') == 'd desc'
        assert h.description('e') == 'e desc'

        h.set('a', 'meta', 'aa')
        h.set('b', 'meta', 'bb')
        h.set('c', 'meta', 'cc')
        h.set('d', 'meta', 'dd')
        h.set('e', 'meta', 'ee')

        assert h.get('a', 'meta') == 'aa'
        assert h.get('b', 'meta') == 'bb'
        assert h.get('c', 'meta') == 'cc'
        assert h.get('d', 'meta') == 'dd'
        assert h.get('e', 'meta') == 'ee'



#!/usr/bin/env python
#
# hierarchy.py - Functions and data structures for working with hierarchical
#                variables.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions and data structures for working with
hierarchical variables.


The :func:`loadHierarchyFile` function will read hierarchy information for
one variable from a text file, and return a :class:`Hierarchy` object.


The :class:`Hierarchy` class allows the hierarchy information about a variable
to be queried.


The hierarchy information for all hierarchical variables in the UKBiobank is
stored in ``funpack/data/hierarchy/`` - these files are available from the
UKBiobank showcase at https://biobank.ctsu.ox.ac.uk/crystal/schema.cgi.
"""


import os.path   as op
import functools as ft

import pandas    as pd
import numpy     as np


HIERARCHY_DATA_NAMES = {
    'icd10' : 19,
    'icd9'  : 87,
    'opcs4' : 240,
    'opcs3' : 259,
}
"""This dictionary contains some UK Biobank hierarchical data codings which
can be looked up by name with the :func:`getHierarchyFilePath` function.
"""


class CircularError(Exception):
    """Error raised by the :meth:`Hierarchy.parents` method in the event
    that a circular relationship is detected in a hierarchy.
    """
    pass


def getHierarchyFilePath(dtable=None,
                         vid=None,
                         name=None,
                         coding=None):
    """Return a path to the file containing hierarchy information for the
    specified variable/name. Pass the path to the :func:`loadHierarchyFile`
    function to load it.

    Hierarchy files can be looked up with one of the following methods, in
    order of precedence:

     1. By specifying a data coding (``coding``). This takes precedence.
     2. By specifying a ``name`` which is present in the
        :attr:`HIERARCHY_DATA_NAMES`.
     3. By passing a :class:`.DataTable` (``dtable``) and variable ID (``vid``)

    Te recognised data type names for use with the second method are listed in
    the :attr:`HIERARCHY_DATA_NAMES` dictionary.

    A :exc:`ValueError` is raised if the variable is unknown, or does not
    have a listed data coding.

    The returned path is not guaranteed to exist.

    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg name:   Data coding name
    :arg coding: Data coding ID
    :returns:    A path to the relevant hierarchical coding file.
    """

    if coding is None:
        if name is not None:
            coding = HIERARCHY_DATA_NAMES[name.lower()]
        elif (dtable is not None) and (vid is not None):
            try:
                coding = int(dtable.vartable.loc[vid, 'DataCoding'])
            except Exception:
                raise ValueError('Variable {} is unknown or does not '
                                 'have a data coding'.format(vid))
    if coding is None:
        raise ValueError('A coding, name, or dtable+vid must be specified')

    hierdir = op.join(op.dirname(__file__), 'data', 'hierarchy')
    return op.join(hierdir, 'coding{}.tsv'.format(int(coding)))


@ft.lru_cache()
def loadHierarchyFile(fname):
    """Reads the hierarchy information for a variable from the given file.

    ``fname`` is assumed to refer to a tab-separated file containing
    the following columns:

      - ``coding``:    A variable value
      - ``meaning``:   Description
      - ``node_id``:   Index in the hierarchy
      - ``parent_id``: Index of this node's parent

    It is assumed that all codings have a sequential ``node_id`` starting from
    1, and that the parent node(s) have a ``parent_id`` of 0.
    The codes are not assumed to be ordered by their ID.

    :arg fname: File containing hierarchy information for one variable
    :returns:   A :class:`Hierarchy` object.
    """

    data      = pd.read_csv(fname, delimiter='\t', index_col=False)
    codings   = data['coding']   .values
    meanings  = data['meaning']  .values
    nodeIds   = data['node_id']  .values
    parentIds = data['parent_id'].values

    order = np.argsort(nodeIds)

    codings   = codings[  order]
    meanings  = meanings[ order]
    nodeIds   = nodeIds[  order] - 1
    parentIds = parentIds[order] - 1

    return Hierarchy(nodeIds, parentIds, codings, meanings)


def codeToNumeric(code, name=None, dtable=None, vid=None, coding=None):
    """Converts a hierarchical code into a numeric version. See the
    :func:`getHierarchyFilePath` for information on the arguments.

    :arg code:   Code to convert
    :arg name:   Data coding name
    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg coding: Data coding ID
    """
    # We use the node IDs defined in
    # the hierarchy file as the
    # numeric version of each coding.
    hier = getHierarchyFilePath(dtable, vid, name, coding)
    hier = loadHierarchyFile(hier)

    try:
        # hierarchy uses 0-indexing,
        # whereas the hierarchy files
        # use 1-indexing
        return int(hier.index(code) + 1)
    except KeyError:
        return np.nan


def numericToCode(code, name=None, dtable=None, vid=None, coding=None):
    """Converts a numeric hierarchical code into its original version

    :arg code:   Code to convert
    :arg name:   Data coding name
    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg coding: Data coding ID
    """
    hier = getHierarchyFilePath(dtable, vid, name, coding)
    hier = loadHierarchyFile(hier)

    try:
        return hier.coding(int(code - 1))
    except KeyError:
        return 'NaN'


class Hierarchy(object):
    """The ``Hierarchy`` class allows information in a hierarchical variable to be
    queried. The :meth:`parents` method will return all parents in the
    hierarchy for a given value (a.k.a. *coding*), and the :meth:`description`
    method will return the description for a value.

    Additional metadata can be added and retrieved for codings via the
    :meth:`set` and :meth:`get` methods.
    """

    def __init__(self, nodes, parents, codings, descs):
        """Create a ``Hierarchy`` object.

        :arg nodes:   Node IDs. Currently assumed to be equivalent to
                      ``np.arange(len(nodes))``
        :arg parents: Parent IDs for each node.
        :arg codings: Value/coding for each node.
        :arg descs:   Description for each node
        """

        self.__nodes    = nodes
        self.__parents  = parents
        self.__codings  = codings
        self.__attrs    = {}
        self.__codeidxs = {}

        self.__attrs['description'] = descs

        for i, c in enumerate(self.__codings):
            self.__codeidxs[c] = i


    @property
    def codings(self):
        """Return a list of all codings in the hierarchy. """
        return list(self.__codings)


    def index(self, coding):
        """Return the node ID for the given ``coding``. """
        return self.__codeidxs[coding]


    def coding(self, nodeID):
        """Return the coding for the given ``nodeID``. """
        return self.__codings[nodeID]


    def parents(self, coding):
        """Return all parents of the given coding. """

        parents = []
        cidx    = self.index(coding)
        pidx    = self.__parents[cidx]
        seen    = set()

        while pidx >= 0:
            if pidx in seen:
                raise CircularError(pidx + 1, str(self.coding(pidx)))
            seen.add(pidx)
            parents.append(self.coding(pidx))
            pidx = self.__parents[pidx]

        return parents


    def description(self, coding):
        """Return the description for the given coding. """
        return self.get(coding, 'description')


    def get(self, coding, attr):
        """Get the given attribue for the given coding. """
        idx = self.index(coding)
        return self.__attrs[attr][idx]


    def set(self, coding, attr, value):
        """Set an attribute for the given coding. """

        if attr not in self.__attrs:
            self.__attrs[attr] = [None] * len(self.__nodes)

        idx = self.index(coding)
        self.__attrs[attr][idx] = value



#!/usr/bin/env python
#
# test_importing_removeSubjects.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import numpy as np

import funpack.importing as importing

from . import gen_DataTable


def test_removeSubjects():

    data = np.random.randint(1, 10, (10, 500))

    # Options in order of precedence:
    #
    #   - exclude:  list of subjects to *exclude*
    #   - exprs:    list of expressions specifying *inclusion*

    # pass through
    dtable = gen_DataTable(data)
    importing.removeSubjects(dtable)
    assert np.all(dtable.index == np.arange(1, 501))
    assert np.all(dtable[:, :] == data.T)

    # exclude
    dtable = gen_DataTable(data)
    importing.removeSubjects(dtable, exclude=[1, 2, 3])
    assert np.all(dtable.index ==         np.arange(4, 501))
    assert np.all(dtable[:, :] == data[:, np.arange(3, 500)].T)

    # expr
    mask     = data[0, :] > 5
    dtable = gen_DataTable(data)
    importing.removeSubjects(dtable, exprs=['v1 > 5'])
    assert np.all(dtable.index == (np.where(mask)[0] + 1))
    assert np.all(dtable[:, :] == data[:, mask].T)

    # expr + exclude
    mask     = data[0, :] > 5
    mask[:3] = 0
    dtable = gen_DataTable(data)
    importing.removeSubjects(dtable, exprs=['v1 > 5'], exclude=[1, 2, 3])
    assert np.all(dtable.index == (np.where(mask)[0] + 1))
    assert np.all(dtable[:, :] == data[:, mask].T)

    # multiple expressions
    dtable = gen_DataTable(data)
    importing.removeSubjects(dtable, exprs=['v1 > 5', 'v2 == 9'])
    mask = (data[0, :] > 5) | (data[1, :] == 9)
    assert np.all(dtable.index == np.where(mask)[0] + 1)
    assert np.all(dtable[:, :] == data[:, mask].T)



#!/usr/bin/env python
#
# test_exporting.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import             warnings
import             os
import textwrap as tw
import os.path  as op

import pandas as pd
import numpy  as np
import           h5py

import funpack.importing      as importing
import funpack.exporting      as exporting
import funpack.exporting_hdf5 as exporting_hdf5
import funpack.datatable      as datatable
import funpack.custom         as custom
import funpack.util           as util

from . import (gen_DataTable,
               clear_plugins,
               patch_base_tables,
               tempdir,
               gen_test_data,
               gen_tables,
               gen_DataTableFromDataFrame)


def test_genColumnNames():

    testdata = np.random.randint(1, 10, (10, 10))

    with patch_base_tables():
        dtable = gen_DataTable(testdata)

    dtable.vartable.loc[6, 'Description'] = 'abcde'

    exp   = {c.name : c.name for c in dtable.allColumns[1:]}
    names = exporting.genColumnNames(dtable)
    assert exp == names

    colpat = '{variable}|{name}|{description}|{visit}|{instance}'
    colmap = { '1-0.0' : 'variable_one', '2-0.0' : 'variable_two'}
    names  = exporting.genColumnNames(dtable, colpat, colmap)

    exp = ['variable_one',
           'variable_two',
           '3|3-0.0||0|0',
           '4|4-0.0||0|0',
           '5|5-0.0||0|0',
           '6|6-0.0|abcde|0|0',
           '7|7-0.0||0|0',
           '8|8-0.0||0|0',
           '9|9-0.0||0|0',
           '10|10-0.0||0|0']

    exp = {c.name : n for c, n in zip(dtable.allColumns[1:], exp)}

    assert exp == names

    exp   = ['var1', 'var2'] + ['{}-0.0'.format(v) for v in range(3, 11)]
    exp   = {c.name : n for c, n in zip(dtable.allColumns[1:], exp)}
    names = exporting.genColumnNames(dtable, None, {'1-0.0' : 'var1',
                                                    '2-0.0' : 'var2'})
    assert exp == names

    exp   = ['{}##0'.format(v) for v in range(1, 11)]
    exp   = {c.name : n for c, n in zip(dtable.allColumns[1:], exp)}
    names = exporting.genColumnNames(dtable, '{variable}##{visit}')
    assert exp == names

    df = pd.DataFrame({'col1' : [1, 2, 3],
                       'col2' : [4, 5, 6],
                       'id'   : [1, 2, 3]}).set_index('id')
    dtable = gen_DataTableFromDataFrame(df)
    exp   = {c.name : c.name for c in dtable.allColumns[1:]}
    names = exporting.genColumnNames(dtable)
    assert exp == names

    exp = ['col1', 'cc2']
    exp   = {c.name : e for c, e in zip(dtable.allColumns[1:], exp)}
    names = exporting.genColumnNames(dtable, None, {'col2' : 'cc2'})
    assert exp == names

    exp = ['00col1', 'cc2']
    exp   = {c.name : e for c, e in zip(dtable.allColumns[1:], exp)}
    names = exporting.genColumnNames(dtable, '{visit}{instance}{name}',
                                     {'col2' : 'cc2'})
    assert exp == names

    exp = ['00col1', '00col2']
    exp   = {c.name : e for c, e in zip(dtable.allColumns[1:], exp)}
    names = exporting.genColumnNames(dtable, '{visit}{instance}{name}')
    assert exp == names



def test_exportData():

    custom.registerBuiltIns()

    testdata = np.random.randint(1, 10, (10, 10)).astype(np.float32)

    def check(gotfile, expected):
        with open(gotfile, 'rt') as f:
            got = f.read().strip()
            return got == expected

    with tempdir():

        dtable = gen_DataTable(testdata)
        exporting.exportData(dtable, 'data.tsv', colpat='{variable}')
        exp = ['\t'.join(['eid'] +
                         ['{}'.format(i) for i in range(1, 11)])] + \
              ['\t'.join([str(i + 1)] + [str(c) for c in r])
               for i, r in enumerate(testdata.T)]
        exp = '\n'.join(exp)
        assert check('data.tsv', exp)

        td = np.copy(testdata)
        td[5, 5] = np.nan
        dtable = gen_DataTable(td)
        exporting.exportData(dtable,
                             'data.tsv',
                             colpat='{variable}',
                             colmap={'1-0.0' : 'var1'},
                             idcol='sub',
                             sep='*',
                             missingValues='boo')
        exp = ['*'.join(['sub', 'var1'] +
                         ['{}'.format(i)  for i in range(2, 11)])] + \
              ['*'.join([str(i + 1)] +
                        ['boo' if np.isnan(c) else str(c) for c in r])
               for i, r in enumerate(td.T)]
        exp = '\n'.join(exp)
        assert check('data.tsv', exp)



def test_exportData_subjid():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(3, 3, 'data.tsv', start_subj=10)

        variables = list(range(0, 4))
        vartable, proctable, cattable, _ = gen_tables(variables)
        colnames  = ['eid'] + ['{}-0.0'.format(v) for v in variables[1:]]
        colobjs   = [datatable.Column(None, c, i, v, 0, 0)
                     for i, (c, v) in enumerate(zip(colnames, variables))]
        data      = pd.read_csv( 'data.tsv', delimiter='\t', index_col=0)

        dtable = datatable.DataTable(
            data, colobjs, vartable, proctable, cattable)

        exporting.exportData(dtable, 'export.tsv')

        got = pd.read_csv('export.tsv', delimiter='\t', index_col=0)

        assert np.all(got.index == [10, 11, 12])


@clear_plugins
def test_exportData_varformats():

    custom.registerBuiltIns()

    def check(gotfile, expected):
        with open(gotfile, 'rt') as f:
            got = f.read().strip()
            return got == expected

    datefmt  = '%Y a %m b %d c'
    timefmt  = '%Y a %m b %d c %M d %H e %S'
    floatfmt = '{} poo'

    @custom.formatter('test_datefmt')
    def datefmt_func(dtable, column, series):
        def format(val):
            return val.strftime(datefmt)
        return series.apply(format)


    @custom.formatter('test_timefmt')
    def timefmt_func(dtable, column, series):
        def format(val):
            return val.strftime(timefmt)
        return series.apply(format)


    @custom.formatter('test_floatfmt')
    def floatfmt_func(dtable, column, series):
        def format(val):
            return floatfmt.format(val)
        return series.apply(format)


    with tempdir():

        gen_test_data(
            3, 3, 'data.tsv', ctypes={1 : 'date', 2 : 'datetime'})

        variables = list(range(1, 4))

        vartable, proctable, cattable, _ = gen_tables(
            variables, {1 : 'Date', 2 : 'Time'})
        colnames  = ['{}-0.0'.format(v) for v in variables]
        colobjs   = [datatable.Column(None, 'eid', 0, 0, 0, 0)] + \
                    [datatable.Column(None, c, i, v, 0, 0)
                     for i, (c, v) in enumerate(zip(colnames, variables), 1)]

        data   = pd.read_csv(
            'data.tsv', delimiter='\t',
            parse_dates=['1-0.0', '2-0.0'],
            index_col=0,
            infer_datetime_format=True)

        dtable = datatable.DataTable(
            data, colobjs, vartable, proctable, cattable)

        expdates  = [v.strftime(datefmt) for v in data['1-0.0']]
        exptimes  = [v.strftime(timefmt) for v in data['2-0.0']]
        expfloats = [ floatfmt.format(v) for v in data['3-0.0']]

        exporting.exportData(dtable,
                             'export.tsv',
                             dateFormat='test_datefmt',
                             timeFormat='test_timefmt',
                             formatters={ 3 : 'test_floatfmt'})

        got = pd.read_csv('export.tsv', delimiter='\t', dtype=str, index_col=0)

        assert np.all(got['1-0.0'] == expdates)
        assert np.all(got['2-0.0'] == exptimes)
        assert np.all(got['3-0.0'] == expfloats)

        expdates = exporting.defaultDateTimeFormat(
            dtable, colobjs[1], dtable[:, '1-0.0'])
        exptimes = exporting.defaultDateTimeFormat(
            dtable, colobjs[2], dtable[:, '2-0.0'])

        exporting.exportData(dtable, 'export.tsv')

        got = pd.read_csv('export.tsv', delimiter='\t', dtype=str, index_col=0)

        assert np.all(got['1-0.0'] == expdates)
        assert np.all(got['2-0.0'] == exptimes)


def test_exportData_TSV_subject_ordering():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)

        exporting.exportData(dtable, 'data.tsv', subjects=[10, 8, 6, 4, 2])
        got = pd.read_csv('data.tsv', delimiter='\t', index_col=0)
        assert np.all(got.index == [10, 8, 6, 4, 2])

        exporting.exportData(dtable,
                             'data.tsv',
                             subjects=[10, 8, 6, 4, 2],
                             numRows=2)
        got = pd.read_csv('data.tsv', delimiter='\t', index_col=0)
        assert np.all(got.index == [10, 8, 6, 4, 2])


def test_exportData_numRows():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)

        exporting.exportData(dtable, 'nochunks.tsv')
        exporting.exportData(dtable, 'chunks.tsv', numRows=2)

        with open('nochunks.tsv', 'rt') as f: nochunks = f.read()
        with open('chunks.tsv',   'rt') as f: chunks   = f.read()

        assert nochunks == chunks


def test_defaultDateTimeFormat():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(3, 3, 'data.tsv', ctypes={1 : 'date', 2 : 'datetime'})

        data = pd.read_csv('data.tsv',
                           delimiter='\t',
                           parse_dates=['1-0.0', '2-0.0'],
                           infer_datetime_format=True,
                           index_col=0)
        dt = gen_DataTableFromDataFrame(data)

        dt.vartable.loc[1, 'Type'] = util.CTYPES.date
        dt.vartable.loc[2, 'Type'] = util.CTYPES.time

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            datecol        = dt[:, '1-0.0']
            datecol[2]     = np.nan
            dt[:, '1-0.0'] = datecol

        def datefmt(d):
            try:    return d.strftime('%Y-%m-%d')
            except: return np.nan
        def timefmt(t):
            try:    return t.strftime('%Y-%m-%dT%H:%M:%S%z')
            except: return np.nan

        expdate = data['1-0.0'].apply(datefmt)
        exptime = data['2-0.0'].apply(timefmt)
        expvar  = data['3-0.0']

        gotdate = exporting.defaultDateTimeFormat(
            dt, dt.columns(1)[0], data['1-0.0'])
        gottime = exporting.defaultDateTimeFormat(
            dt, dt.columns(2)[0], data['2-0.0'])
        gotvar  = exporting.defaultDateTimeFormat(
            dt, dt.columns(3)[0], data['3-0.0'])

        expdna = expdate.isna()
        exptna = exptime.isna()

        assert np.all(expdate.isna() == expdna)
        assert np.all(exptime.isna() == exptna)

        assert np.all(expdate[~expdna] == gotdate[~expdna])
        assert np.all(exptime[~exptna] == gottime[~exptna])
        assert np.all(expvar           == gotvar)


def test_exportHDF5():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(5, 10, 'data.tsv', ctypes={1 : 'date', 2 : 'datetime'})

        colnames = ['one', 'two', 'three', 'four', 'five']

        data = pd.read_csv('data.tsv',
                           delimiter='\t',
                           parse_dates=['1-0.0', '2-0.0'],
                           infer_datetime_format=True,
                           index_col=0)
        dt = gen_DataTableFromDataFrame(data)

        exporting_hdf5.exportHDF5(dt,
                                  'out_funpack.h5',
                                  list(dt.index),
                                  'eid',
                                  colnames,
                                  key='h5key',
                                  style='funpack')
        exporting_hdf5.exportHDF5(dt,
                                  'out_pandas.h5',
                                  list(dt.index),
                                  'eid',
                                  colnames,
                                  key='h5key',
                                  style='pandas')

        exp    = dt[:, :]
        gotpd  = pd.read_hdf('out_pandas.h5')
        gotukb = h5py.File('out_funpack.h5', 'r')
        assert gotpd.index.name     == 'eid'
        assert np.all(gotpd.columns == colnames)

        for i, c in enumerate(exp.columns):
            assert np.all(gotpd[colnames[i]] == exp[c])

        gotidx = gotukb['h5key/eid']

        assert np.all(exp.index[:] == gotidx)

        for i, col in enumerate(dt.allColumns[1:]):
            coldata = gotukb['h5key/{}'.format(colnames[i])]

            if col.name in ('1-0.0', '2-0.0'):
                expcol = exporting.defaultDateTimeFormat(
                    dt, col, exp[col.name])
            else:
                expcol = exp[col.name]

            assert np.all(np.asarray(coldata) == expcol)


def test_exportData_HDF_subject_ordering():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)

        # pandas style + no chunking
        exporting.exportData(dtable,
                             'data.h5',
                             fileFormat='hdf5',
                             subjects=[10, 8, 6, 4, 2])
        got  = pd.read_hdf('data.h5')
        assert np.all(got.index == [10, 8, 6, 4, 2])

        # pandas style + chunking
        exporting.exportData(dtable,
                             'data.h5',
                             fileFormat='hdf5',
                             subjects=[10, 8, 6, 4, 2],
                             numRows=2)
        got  = pd.read_hdf('data.h5')
        assert np.all(got.index == [10, 8, 6, 4, 2])

        # funpack style
        exporting.exportData(dtable,
                             'data.h5',
                             style='funpack',
                             fileFormat='hdf5',
                             subjects=[10, 8, 6, 4, 2])

        got = exporting_hdf5.importFunpackStyle(
            'data.h5', 'eid', key='funpack')

        assert np.all(got.index == [10, 8, 6, 4, 2])



def test_compound():

    data = tw.dedent("""
    "eid","1-0.0","2-0.0"
    1,"1,2,3",8
    2,"4,5,6",8
    3,"7,8,9",8
    """).strip()

    exp = tw.dedent("""
    eid\t1-0.0\t2-0.0
    1\t1.0,2.0,3.0\t8
    2\t4.0,5.0,6.0\t8
    3\t7.0,8.0,9.0\t8
    """).strip()


    def parseCompound(v):
        return np.fromstring(v, sep=',')

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1, 2])
        dt, _ = importing.importData('data.txt', vartable, proctable, cattable)

        dt[:, '1-0.0'] = dt[:, '1-0.0'].apply(parseCompound)

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv',
                             formatters={1 : 'compound'})

        with open('out.txt', 'rt') as f:
            got = f.read().strip()

        assert got == exp


def test_exporting_id_column():
    data = tw.dedent("""
    my_id,col1
    1,11
    2,12
    3,13
    """).strip()

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1], datafiles=['data.txt'])
        dt, _ = importing.importData('data.txt', vartable, proctable, cattable)

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv')


        got = pd.read_csv('out.txt', delimiter='\t', index_col=0)
        assert got.index.name == 'my_id'

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv',
                             idcol='my_id_renamed')
        got = pd.read_csv('out.txt', delimiter='\t', index_col=0)
        assert got.index.name == 'my_id_renamed'


def test_exporting_no_data():

    data = tw.dedent("""
    eid,1-0.0,2-0.0
    1,11,21
    2,12,22
    3,13,23
    4,14,24
    5,15,25
    """)

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1])
        dt, _ = importing.importData(
            'data.txt', vartable, proctable, cattable, variables=[3])

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv')

        exp = "eid\n1\n2\n3\n4\n5"
        got = open('out.txt').read().strip()
        assert exp == got


def test_exporting_non_numeric():

    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0,4-0.0
    1,11,a,31,ff
    2,12,b,32,gg
    3,13,c,33,hh
    4,14,d,34,ii
    5,15,e,35,jj
    """)
    expn = tw.dedent("""
    eid\t1-0.0\t3-0.0
    1\t11\t31
    2\t12\t32
    3\t13\t33
    4\t14\t34
    5\t15\t35
    """)
    expnn = tw.dedent("""
    eid\t2-0.0\t4-0.0
    1\ta\tff
    2\tb\tgg
    3\tc\thh
    4\td\tii
    5\te\tjj
    """)

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables(range(1, 5))
        dt, _ = importing.importData(
            'data.txt', vartable, proctable, cattable)

        exporting.exportData(dt,
                             'numerics.txt',
                             nonNumericFile='non_numerics.txt',
                             fileFormat='tsv')

        gotn = open('numerics.txt')    .read().strip()
        gotnn = open('non_numerics.txt').read().strip()

        assert expn .strip() == gotn .strip()
        assert expnn.strip() == gotnn.strip()

        # if no non-numeric columns,
        # file should not be created
        os.remove('non_numerics.txt')
        with open('data.txt', 'wt') as f:
            f.write(expn)
        vartable, proctable, cattable, _ = gen_tables([1, 3])
        dt, _ = importing.importData(
            'data.txt', vartable, proctable, cattable)
        exporting.exportData(dt,
                             'numerics.txt',
                             nonNumericFile='non_numerics.txt',
                             fileFormat='tsv')
        gotn = open('numerics.txt').read().strip()

        assert gotn == expn.strip()
        assert not op.exists('non_numerics.txt')



#!/usr/bin/env python
#
# merging.py - Merging data from multiple input files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions which can merge data from multiple input
files.


.. autosummary::
   :nosignatures:

   mergeDataFrames

"""


import itertools as it
import              logging
import              collections

import numpy     as np
import pandas    as pd


log = logging.getLogger(__name__)


def mergeDataFrames(data, cols, axis, strategy, dryrun=False):
    """Merges one or more ``pandas.DataFrames`` according to the given
    ``axis`` and ``strategy``.

    :arg data:     List of ``DataFrame`` objects to merge.

    :arg cols:     List of lists - :class:`.Column` objects representing the
                   columns in each data set in ``data``.

    :arg axis:     Axis to merge on - either ``subjects`` or ``variables``.

    :arg strategy: Strategy to use for merging ``data``, either ``union`` (an
                   outer join), ``intersection`` (inner join), or ``naive``
                   (naive concatenation along ``axis``).

    :arg dryrun:   If ``True``, only ``cols`` is merged.

    :returns:  A tuple containing:

                 - a new ``DataFrame`` containing the merged ``data``, or
                   ``None`` if ``dryrun is True``.

                 - A list of :class:`.Column` objects representing the columns
                   that were kept. The index column is at the beginning of the
                   list.

    .. warning:: A dry run may produce different results with the naive merge
                 strategy.
    """

    if axis not in (0, '0', 'rows', 'subjects',
                    1, '1', 'cols', 'columns', 'variables'):
        raise ValueError('Invalid axis: {}'.format(axis))

    if strategy not in ('naive', 'union', 'intersection', 'inner', 'outer'):
        raise ValueError('Invalid merge strategy: {}'.format(strategy))

    axis     = {0              : 0,
                '0'            : 0,
                'rows'         : 0,
                'subjects'     : 0,
                1              : 1,
                '1'            : 1,
                'cols'         : 1,
                'columns'      : 1,
                'variables'    : 1}[axis]
    strategy = {'naive'        : 'naive',
                'inner'        : 'inner',
                'intersection' : 'inner',
                'outer'        : 'outer',
                'union'        : 'outer'}[strategy]

    if len(data) == 0:
        raise ValueError('No data!')

    naive = strategy == 'naive'

    if strategy == 'inner': join = 'inner'
    else:                   join = 'outer'

    # Separate out the index and non-
    # index columns for each data frame
    idxcols  = []
    datacols = [[] for _ in cols]
    for dfi, dfcols in enumerate(cols):
        for col in dfcols:
            if col.vid == 0: idxcols.append(col)
            else:            datacols[dfi].append(col)

    # Only one file - no merging required
    if len(data) == 1:
        return data[0], [idxcols[0]] + datacols[0]

    # Build a list of Column objects
    # which describe what the merged
    # data frame will look like.
    # This list will be assigned
    # back to the cols variable.

    # naive concatenation.
    if naive:

        # if performing a dry run, we can't look
        # at the data dimensions, so we take the
        # first file as the definitive one.
        if dryrun:
            if axis == 0: cols = cols[0]
            else:         cols = cols[0] + list(it.chain(*datacols[1:]))

        # If concatenating rows, we assume that
        # the columns are aligned in each file.
        # We take the column names from the file
        # with the most columns.
        elif axis == 0:
            lens = [len(d.columns) for d in data]
            rows = list(it.chain(*[d.index for d in data]))
            cols = cols[lens.index(max(lens))]

        # If concatenating columns, we assume
        # that the subjects are aligned in each
        # file. We take the subject IDs from the
        # file with the largest number of rows.
        #
        # We re-generate the list of expected
        # rows and columns in the data, dropping
        # the index column
        else:
            lens   = [len(d.index) for d in data]
            maxi   = lens.index(max(lens))
            idxcol = idxcols[maxi]
            rows   = data[maxi].index
            cols   = [idxcol] + list(it.chain(*datacols))

    # concatenate rows with inner join - only
    # retain the columns that are present in
    # every dataframe. Identify the Column
    # objectsd which represent these retained
    # columns.
    elif axis == 0 and strategy == 'inner':
        idxcol    = idxcols[0]
        allcols   = list(it.chain(*datacols))
        colnames  = [[c.name for c in dc] for dc in datacols]
        innercols = set.intersection(*[set(cn) for cn in colnames])
        colnames  = list(it.chain(*colnames))
        colidxs   = [colnames.index(ic) for ic in innercols]
        cols      = [idxcol] + [allcols[i] for i in sorted(colidxs)]

    # concatenate rows with outer join - all
    # unique column names will be retained.
    elif axis == 0 and strategy == 'outer':
        idxcol    = idxcols[0]
        uniqnames = []
        uniqcols  = []

        for col in it.chain(*datacols):
            if col.name not in uniqnames:
                uniqnames.append(col.name)
                uniqcols .append(col)

        cols = [idxcol] + uniqcols

    # concatenate columns (horizontally) -
    # all columns are retained even if
    # there are duplicate column names
    else:
        idxcol = idxcols[0]
        cols   = [idxcol] + list(it.chain(*datacols))

    if dryrun:
        return None, cols

    log.debug('Merging %u data sets (axis: %s, strategy: %s)',
              len(data), axis, strategy)

    # We have to reset row/column labels,
    # otherwise pandas will try to do a
    # real join.
    if naive:
        for d in data:
            d.columns = range(len(d.columns))
            d.reset_index(drop=True, inplace=True)

    merged = pd.concat(data,
                       axis=axis,
                       join=join,
                       ignore_index=naive,
                       sort=True,
                       copy=False)

    if len(merged) == 0:
        log.warning('Merged dataframe is empty! Are '
                    'index column positions correct?')

    # Re-label subjects and variables,
    # make sure that dataframe column
    # order matches cols list order.
    if naive:
        merged.index   = rows
        merged.columns = [c.name for c in cols[1:]]

    # If the input data frames
    # have different index names,
    # the merged index will not
    # have a name. So make sure
    # it has a name.
    merged.index.name = cols[0].name

    # Warn if we concatenated rows, and
    # there are duplicate subject IDs.
    if axis == 0:
        idxs, counts = np.unique(merged.index, return_counts=True)
        notuniq      = counts > 1
        if np.any(notuniq):
            log.warning('Duplicate subject IDs in data: %s',
                        idxs[notuniq][:5])

    # Or if we concatenated columns, and
    # there are duplicate column names
    # or VIDs
    else:
        # names
        counts = collections.Counter(merged.columns)
        if any([c > 1 for c in counts.values()]):
            log.warning('Duplicate column names in data: %s',
                        [n for n, c in counts.items() if c > 1])

        # vids
        allcols = [(c.vid, c.visit, c.instance) for c in cols]
        counts  = collections.Counter(allcols)
        if any([c > 1 for c in counts.values()]):
            log.warning('Duplicate column VIDs in data: %s',
                        [n for n, c in counts.items() if c > 1])

    return merged, cols



#!/usr/bin/env python

import itertools as it

from funpack import processor

# Processor functions are passed:
#
#   - dtable: An object which provides access to the data set.
#   - vid:    The variable ID of the column(s) to be cleaned.
#
# Processor functions can do any of the following:
#
#   - modify existing columns in place,
#   - return a list of columns that should be removed
#   - return a list of columns that should be added
@processor()
def sum_squares(dtable, vids):

    cols    = it.chain(*[dtable.columns(v) for v in vids])
    series  = [dtable[:, c.name] for c in cols]
    squares = [s * s for s in series]
    sumsq   = sum(squares)

    sumsq.name = 'sum_square({})'.format(','.join([str(v) for v in vids]))

    # The value returned by a processor function differs
    # depending on what it wishes to do. In this case,
    # we are returning a list of new pandas.Series to be
    # added as columns, and a list of integer variable
    # IDs, one for each new column. The variable IDs are
    # optional, so we are just returning None instead.
    return [sumsq], None



#!/usr/bin/env python
#
# exporting_tsv.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import logging

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from . import util
from . import custom


log = logging.getLogger(__name__)


TSV_SEP = '\t'
"""Default separator string to use in TSV-style output files."""


@custom.exporter('tsv')
def exportTSV(dtable,
              outfile,
              subjects,
              idcol,
              colnames,
              sep=None,
              missingValues=None,
              dateFormat=None,
              timeFormat=None,
              formatters=None,
              numRows=None,
              nonNumericFile=None,
              **kwargs):
    """Export data to a TSV-style file.

    :arg dtable:         :class:`.DataTable` containing the data

    :arg outfile:        File to output to

    :arg subjects:       Sequence containing subjects (and order) to export.

    :arg idcol:          Name to use for the subject ID column

    :arg colnames:       Sequence containing column names

    :arg sep:            Separator character to use. Defaults to
                         :attr:`TSV_SEP`

    :arg missingValues:  String to use for missing/NA values. Defaults to the
                         empty string.

    :arg dateFormat:     Name of formatter to use for date columns.

    :arg timeFormat:     Name of formatter to use for time columns.

    :arg formatters:     Dict of ``{ [vid|column] : formatter }`` mappings,
                         specifying custom formatters to use for specific
                         variables.

    :arg numRows:        Number of rows to write at a time. Defaults to writing
                         all rows in one go.

    :arg nonNumericFile: If provided, non-numeric columns (after formatting)
                         are saved to this file instead of to ``outfile``
    """
    if sep           is None: sep           = TSV_SEP
    if missingValues is None: missingValues = ''
    if dateFormat    is None: dateFormat    = 'default'
    if timeFormat    is None: timeFormat    = 'default'
    if formatters    is None: formatters    = {}
    if numRows       is None: numRows       = len(dtable)
    if colnames      is None: colnames      = True

    nchunks   = int(np.ceil(len(subjects) / numRows))
    vartable  = dtable.vartable

    # Created during the first iteration.
    # If nonNumericFile is specified, we
    # store the names of all numeric and
    # non-numeric columns here so we can
    # figure out which columns to put
    # where.
    numericCols    = None
    nonNumericCols = None

    log.info('Writing %u columns in %u chunk(s) to %s ...',
             len(dtable.allColumns), nchunks, outfile)

    for chunki in range(nchunks):

        cstart  = chunki * numRows
        cend    = cstart + numRows
        csubjs  = subjects[cstart:cend]
        towrite = pd.DataFrame(index=csubjs)

        for col in dtable.allColumns:

            vid = col.vid

            if vid == 0:
                continue

            name      = col.name
            series    = dtable[csubjs, name]
            formatter = formatters.get(vid, None)

            if vid in vartable.index: vtype = vartable['Type'][vid]
            else:                     vtype = None

            # allow formatters to be
            # specified by column name
            # as well
            if formatter is None:
                formatter = formatters.get(name, None)

            # fall back to date/time formatting
            # if relevant for this column
            if formatter is None:
                if   vtype == util.CTYPES.date:
                    formatter = dateFormat
                elif vtype == util.CTYPES.time or \
                     pdtypes.is_datetime64_any_dtype(series):
                    formatter = timeFormat

            if formatter is not None:
                log.debug('Formatting column %s [chunk %u] with %s formatter',
                          name, chunki, formatter)
                towrite[name] = custom.runFormatter(
                    formatter, dtable, col, series)
            else:
                towrite[name] = series

        if nonNumericFile is None:
            numericCols     = colnames
            nonNumericCols  = None
            numericChunk    = towrite
            nonNumericChunk = None
        else:
            if numericCols is None:
                numericCols    = [c for c in towrite.columns
                                  if     pdtypes.is_numeric_dtype(towrite[c])]
                nonNumericCols = [c for c in towrite.columns
                                  if not pdtypes.is_numeric_dtype(towrite[c])]

                if len(nonNumericCols) > 0:
                    log.debug('Redirecting %i non-numeric columns to %s '
                              '(remaining %i columns will be written to %s)',
                              len(nonNumericCols), nonNumericFile,
                              len(numericCols),    outfile)
                else:
                    log.debug('No non-numeric columns present - not creating '
                              '%s', nonNumericFile)
                    nonNumericFile  = None
                    nonNumericCols  = None
                    nonNumericChunk = None

            numericChunk = towrite[numericCols]
            if nonNumericCols is not None:
                nonNumericChunk = towrite[nonNumericCols]

        _writeChunk(numericChunk,
                    chunki,
                    outfile,
                    sep,
                    missingValues,
                    numericCols,
                    idcol)
        if nonNumericFile is not None:
            _writeChunk(nonNumericChunk,
                        chunki,
                        nonNumericFile,
                        sep,
                        missingValues,
                        nonNumericCols,
                        idcol)


def _writeChunk(chunk, i, outfile, sep, missingValues, colnames, idcol):
    """Write a chunk of data to a file.

    :arg chunk:         ``pandas.DataFrame`` containing the data to write
    :arg i:             Chunk index - if ``0``, column headers are written
    :arg outfile:       File to write to
    :arg sep:           Separater character
    :arg missingValues: String to use for missing values
    :arg colnames:      Names for all columns
    :arg idcol:         Name for the index column
    """

    if i > 0:
        mode     = 'a'
        colnames = False
        idcol    = None
    else:
        mode = 'w'

    chunk.to_csv(outfile,
                 sep=sep,
                 na_rep=missingValues,
                 header=colnames,
                 index_label=idcol,
                 mode=mode)



#!/usr/bin/env python
#
# conftest.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import inspect

import funpack.fileinfo  as fileinfo
import funpack.hierarchy as hierarchy
import funpack.coding    as coding


def fake_cache_clear():
    pass


fileinfo.sniff              = inspect.unwrap(fileinfo.sniff)
fileinfo.fileinfo           = inspect.unwrap(fileinfo.fileinfo)
hierarchy.loadHierarchyFile = inspect.unwrap(hierarchy.loadHierarchyFile)
coding.loadCodingFile       = inspect.unwrap(coding.loadCodingFile)

fileinfo.sniff             .cache_clear = fake_cache_clear
fileinfo.fileinfo          .cache_clear = fake_cache_clear
hierarchy.loadHierarchyFile.cache_clear = fake_cache_clear
coding.loadCodingFile      .cache_clear = fake_cache_clear



#!/usr/bin/env python
#
# icd10.py - Query the ICD10 disease coding hierarchy.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for working with the `ICD10
<https://en.wikipedia.org/wiki/ICD-10>`_ disease coding hierarcy.


The :func:`codeToNumeric` function will take an ICD10 coding, and return
a numeric variant of it.


The :func:`storeCodes` function allows sets of ICD10 codes to be stored
so that they can be saved out to a file via the :func:`saveCodes` function, at
a later stage.
"""


import              logging
import itertools as it
import functools as ft

import numpy     as np
import pandas    as pd

from . import util
from . import hierarchy


log = logging.getLogger(__name__)


@util.deprecated('Use funpack.hierarchy.codeToNumeric instead')
def codeToNumeric(code):
    """Converts an ICD10 code into a numeric version. """
    return hierarchy.codeToNumeric(code, 'icd10')


@util.deprecated('Use funpack.hierarchy.numericToCode instead')
def numericToCode(code):
    """Converts a numeric ICD10 code into its original version. """
    return hierarchy.numericToCode(code, 'icd10')


def storeCodes(codes):
    """Stores the given sequence of ICD10 codes, so they can be exported to
    file at a later stage.

    The codes are stored in a list called ``store``, an attribute of this
    function.

    :arg codes: Sequence of ICD10 codes to add to the mapping file
    """
    store = getattr(storeCodes, 'store', [])
    store.append(codes)
    storeCodes.store = store


def saveCodes(fname, hier, fields=None):
    """Saves any codes which have been stored via :func:`storeCodes` out to
    the specified file.

    :arg fname:     File to save the codes to.

    :arg hier:      :class:`.Hierarchy` object containing the ICD10
                    hierarchy information.

    :arg fields:    Sequence of fields to include in the ``mapfile``. Defaults
                    to ``['code', 'value', 'description', 'parent_descs]``. May
                    contain any of the following:
                      - ``'code'``
                      - ``'value'``
                      - ``'description'``
                      - ``'parent_codes'``
                      - ``'parent_descs'``
    """

    if fields is None:
        fields = ['code', 'value', 'description', 'parent_descs']

    valid = ['code', 'value', 'description', 'parent_codes', 'parent_descs']
    if not all([f in valid for f in fields]):
        raise ValueError('Invalid field in: {}'.format(fields))

    store = getattr(storeCodes, 'store', [])
    store = pd.Series(list(it.chain(*store)))
    store = store[store.notna()]
    codes = np.sort(store.unique())

    def parent_codes(c):
        return ','.join(reversed(hier.parents(c)))

    def parent_descs(c):
        parents = reversed(hier.parents(c))
        descs   = [hier.description(p) for p in parents]
        return ' '.join(['[{}]'.format(d) for d in descs])

    df = pd.DataFrame({'code' : codes})

    for f in fields:
        if   f == 'code':         continue
        elif f == 'value':        func = ft.partial(hierarchy.codeToNumeric,
                                                    name='icd10')
        elif f == 'description':  func = hier.description
        elif f == 'parent_codes': func = parent_codes
        elif f == 'parent_descs': func = parent_descs

        df[f] = df['code'].apply(func)

    log.debug('Saving %u ICD10 codes to %s', len(df), fname)

    df = df[fields]
    df.to_csv(fname, sep='\t', index=False)



#!/usr/bin/env python
#
# test_custom.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import textwrap as tw

import pytest

import funpack.custom as custom

from . import tempdir, clear_plugins


@clear_plugins
def test_register():

    ran = {}

    custom.registry.clearRegistry()

    @custom.loader('test_loader')
    def custom_loader():
        ran['loader'] = True

    @custom.sniffer('test_sniffer')
    def custom_sniffer():
        ran['sniffer'] = True

    @custom.formatter('test_formatter')
    def custom_formatter():
        ran['formatter'] = True

    @custom.cleaner('test_cleaner')
    def custom_cleaner():
        ran['cleaner'] = True

    @custom.processor('test_processor')
    def custom_processor():
        ran['processor'] = True

    @custom.exporter()
    def exporter1():
        ran['exporter1'] = True

    @custom.registerPlugin('exporter', 'test_exporter2')
    def exporter2():
        ran['exporter2'] = True

    with pytest.raises(ValueError):
        @custom.registerPlugin('bad_type')
        def bad_plugin():
            pass

    assert custom.listPlugins('loader')    == ['test_loader']
    assert custom.listPlugins('sniffer')   == ['test_sniffer']
    assert custom.listPlugins('formatter') == ['test_formatter']
    assert custom.listPlugins('cleaner')   == ['test_cleaner']
    assert custom.listPlugins('processor') == ['test_processor']
    assert custom.listPlugins('exporter')  == ['exporter1', 'test_exporter2']

    assert custom.exists('loader',    'test_loader')
    assert custom.exists('sniffer',   'test_sniffer')
    assert custom.exists('formatter', 'test_formatter')
    assert custom.exists('cleaner',   'test_cleaner')
    assert custom.exists('processor', 'test_processor')
    assert custom.exists('exporter',  'exporter1')
    assert custom.exists('exporter',  'test_exporter2')

    custom.run('loader',    'test_loader')
    custom.run('sniffer',   'test_sniffer')
    custom.run('formatter', 'test_formatter')
    custom.run('cleaner',   'test_cleaner')
    custom.run('processor', 'test_processor')
    custom.run('exporter',  'exporter1')
    custom.run('exporter',  'test_exporter2')

    assert ran['loader']
    assert ran['sniffer']
    assert ran['formatter']
    assert ran['cleaner']
    assert ran['processor']
    assert ran['exporter1']
    assert ran['exporter2']


@clear_plugins
def test_loadPluginFile():

    contents = tw.dedent("""

    import funpack.custom as custom

    @custom.cleaner('poo')
    def poo():
        pass
    """).strip()

    with tempdir():

        with open('plugin.py', 'wt') as f:
            f.write(contents)

        custom.loadPluginFile('plugin.py')
        assert custom.exists('cleaner', 'poo')

    # also load a built-in plugin file
    custom.loadPluginFile('fmrib.py')
    assert custom.exists('formatter', 'FMRIBImagingDate')



#!/usr/bin/env python
#
# test_processing_chain.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap as tw

import numpy as np

import funpack.main as main


from . import tempdir, patch_logging


@patch_logging
def test_binariseCategorical_outputs_go_through_sparsity_check():

    data = tw.dedent("""
    eid\t1-0.0\t2-0.0
    1\t0\t1
    2\t1\t1
    3\t1\t1
    4\t1\t2
    5\t1\t2
    6\t1\t2
    7\t1\t3
    8\t1\t3
    9\t1\t3
    10\t1\t3
    """).strip()

    # binarised outputs from var 1
    # should fail the maxcat test,
    # but those from var 2 should
    # pass

    exp = tw.dedent("""
    eid\t2-1\t2-2\t2-3
    1\t1\t0\t0
    2\t1\t0\t0
    3\t1\t0\t0
    4\t0\t1\t0
    5\t0\t1\t0
    6\t0\t1\t0
    7\t0\t0\t1
    8\t0\t0\t1
    9\t0\t0\t1
    10\t0\t0\t1
    """).strip()

    variables = tw.dedent("""
    ID\tType\tDescription\tDataCoding\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean
    1\tInteger
    2\tInteger
    """).strip()

    with tempdir():

        with open('data.tsv',      'wt') as f: f.write(data)
        with open('variables.tsv', 'wt') as f: f.write(variables)

        main.main('-sp -n -n '
                  '-vf variables.tsv '
                  '-apr 1 binariseCategorical(nameFormat="{vid}-{value}") '
                  '-apr 2 binariseCategorical(nameFormat="{vid}-{value}") '
                  '-apr all removeIfSparse(maxcat=0.8,abscat=False) '
                  'out.tsv data.tsv'.split())

        with open('out.tsv', 'rt') as f:
            got = f.read().strip()

    assert got == exp



#!/usr/bin/env python
#
# expression.py - Parser for ParentValue expressions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for parsing conditional and logical
expressions, and the :class:`Expression` class for representing a parsed
expression.


.. autosummary::
   :nosignatures:

   Expression
   parseExpression
   variablesInExpression
   calculateExpressionEvaluationOrder


For a given variable, the ``ParentValues`` column of the variable table may
contain one or more *expressions*, which define conditions that parent
variables of the variable may meet in order for the variable value to be
replaced. This module contains logic for parsing and evaluating a single
expression - the evaluation of multiple comma-separated expressions is handled
in the :mod:`.importing` module.


An *expression* comprises one or more *conditional statements* (or
*statements* for short). A statement has the form::

    variable operator value

where:

  - ``variable`` is the ID of a parent variable of the variable in question.
    Variable IDs must be an integer preceded by the letter ``v``.
  - ``operator`` is a comparison operator (e.g. equals, greater than, etc.).
  - ``value`` is either ``'na'`` indicating missing, or a numeric value against
    which the parent variable is to be compared.


The following comparison operators are allowed (and the symbols used in a
statement can be found in the :attr:`SYMBOLS` dictionary):

 - equal to
 - not equal to
 - greater than
 - greater than or equal to
 - less than
 - less than or equal to


The *equal to* and *not equal to* operators may be used with a value of
``'na'`` to test whether the values for a variable are missing or present
respectively.


Multiple conditional statements may be combined with ``and``, ``or``, and
``not`` logical operations (specific symbols can be found in the
:attr:`SYMBOLS` dictionary), and precedence may be enforced with the use of
round brackets.
"""


import              logging
import              collections
import itertools as it
import functools as ft
import pyparsing as pp


log = logging.getLogger(__name__)


SYMBOLS = {
    'var' : 'v',
    'and' : '&&',
    'or'  : '||',
    'not' : '~',
    'eq'  : '==',
    'ne'  : '!=',
    'lt'  : '<',
    'le'  : '<=',
    'gt'  : '>',
    'ge'  : '>=',
    'na'  : 'na',
}
"""This dictionary contains the symbols for variables and operations that
may be used in expressions.
"""


class Expression(object):
    """The ``Expression`` class is a convenience class which can be used to
    parse and access an expression.
    """


    def __init__(self, expr):
        """Create an ``Expression`` object from the string ``expr``.

        :arg expr: Expression to be parsed.
        """

        self.__variables  = None
        self.__origExpr   = expr
        self.__expression = parseExpression(expr)


    def __str__(self):
        """Return the original string representation of the expression. """
        return self.__origExpr


    def __repr__(self):
        """Return the original string representation of the expression. """
        return str(self)


    @property
    def variables(self):
        """Return a list of all variables used in the expression. """

        if self.__variables is None:
            self.__variables = variablesInExpression(self.__expression)
        return self.__variables


    def evaluate(self, dtable, data):
        """Evaluates this ``Expression`` and returns the result.

        :arg dtable: The :class:`.DataTable` containing the data.

        :arg data:   Dictionary  containing ``{ variable : column_name }``
                     mappings from the variables used in the expressions to
                     columns in ``dtable``.

        :returns:    The outcome of the expression - ``True`` or ``False``.
        """
        return self.__expression(dtable, data)


def calculateExpressionEvaluationOrder(vids, exprs):
    """Identifies hierarchical relationships between variables.

    Given the variable table, identifies the hierarchical relationship order
    between all variables, and all parent variables used within their
    expressions.

    :arg vids:  Sequence of variable IDs

    :arg exprs: Sequence of parsed expression functions (as returned by
                :func:`parseExpression`), one for each variable in
                ``variables``. For each variable, there may be either one
                expression function, or a sequence of them.

    :returns:   A list of tuples, each containing:
                 - A hierarchy level
                 - A list of all variables at that level
                The list is in ascending order, by the hierarchy level
    """

    if len(vids) != len(exprs):
        raise ValueError('vids/exprs lengths don\'t match')

    # get a list of parents for each
    # var, then turn this into a dictionary
    # of { var : parents } mappings
    parents = []
    for expr in exprs:
        if not isinstance(expr, collections.Sequence):
            expr = [expr]
        parents.append(list(it.chain(*[e.variables for e in expr])))
    children = {i : list(sorted(p)) for i, p in zip(vids, parents)}

    # Make a list of all child/
    # parent variable IDs.
    childvids = vids
    allvids   = sorted(set(it.chain(vids, *parents)))

    # Then create a dictionary which will
    # store a hierarchy level for each
    # variable, where zero indicates that
    # the variable has no dependants.
    levels = collections.OrderedDict([(i, 0) for i in allvids])

    # Determine the hierarchy levels. For
    # each variable, we set the level on
    # each of its parents to one plus its
    # level. We use the seen set to keep
    # track of variables that have already
    # been visit, and thus to detect
    # circular or self-dependencies.
    def update(vid, level, seen):

        if vid in seen:
            raise ValueError('Circular dependency identified: {}'.format(vid))

        seen.add(vid)
        levels[vid] = level

        for parent in children.get(vid, set()):

            # only update parent level if it needs to
            # be updated - it may have already been
            # set by a sibling of this variable (i.e.
            # another variable at the same hierarchy
            # level as this one).
            if levels[parent] <= level:
                update(parent, level + 1, seen)

    for vid, level in levels.items():
        update(vid, level, set())

    # Now we can just sort the variables
    # by hierarchy to get the expression
    # evaluation order.
    bylevel = collections.OrderedDict()
    for vid, level in levels.items():
        if vid not in childvids:
            continue
        if level not in bylevel: bylevel[level]     = [vid]
        else:                    bylevel[level].append(vid)

    return list(sorted(bylevel.items(), key=lambda l: l[0]))


def parseExpression(expr):
    """Parses a string containing an expression.

    The expression may contain conditional statements of the form::
        variable comparison_operator value

    combined with logical expressions using symbols for ``and``, ``or``, and
    ``not``.

    The ``parseExpression`` function, given an expression string, will return
    a function that can be used to evaluate the expression. An expression
    function expects to be given two arguments:

    - A ``pandas.DataFrame`` which contains the data on all variables used
      in the expression
    - A dictionary containing ``{variable : column}`` mappings from the
      variables used in the expression to the columns of the data frame.

    An expression function will simply return ``True`` or ``False``, depending
    on the outcome of the expression.

    Expression functions have a few attributes containing metadata about the
    expression:

      - ``ftype`` contains the expression type, either ``logical_not``,
        ``logical`` (for *and*/*or* operations), or ``condition`` (for
        comparison operations)
      - ``operation`` contains the operation symbol

    Boolean *and*/*or* functions contain ``operand1`` and ``operand2``
    attributes which refer to the expression functions they will be applied
    to. Similarly, boolean *not* functions contain an ``operand`` attribute
    which refers to the expression function it will be applied to.  Comparison
    expression functions contain ``variable`` and ``value`` attributes, which
    contain the variable name and the value involved in the comparison.

    :arg expr: String containing an expression.
    :returns:  A function which can be used to evaluate the expression.
    """
    try:
        return list(makeParser().parseString(expr, parseAll=True))[0]
    except Exception as e:
        log.error('Error parsing expression "{}": {}'.format(expr, e))
        raise e


def variablesInExpression(expr):
    """Given an expression returned by :func:`parseExpression`, extracts all
    variables used in the expression.

    :arg expr: A *parsed* expression, produced by :func:`parseExpression`.
    :returns:  A set containing all of the variables that are mentioned in
               the expression.
    """

    if expr.ftype == 'condition':
        return set([expr.variable])

    elif expr.ftype == 'logical':
        variables = set()
        variables.update(variablesInExpression(expr.operand1))
        variables.update(variablesInExpression(expr.operand2))
        return variables

    elif expr.ftype == 'logical_not':
        return variablesInExpression(expr.operand)


def makeParser():
    """Generates a ``pyparsing`` parser which can be used to parse expressions.

    :returns: A ``pyparsing`` object which can parse an expression.
    """

    if getattr(makeParser, 'parser', None) is not None:
        return makeParser.parser

    CMP     = ['eq', 'ne', 'lt', 'le', 'gt', 'ge']
    CMP     = pp.oneOf([SYMBOLS[c] for c in CMP])
    EQS     = pp.oneOf([SYMBOLS[c] for c in ['eq', 'ne']])
    AND     = pp.CaselessLiteral(SYMBOLS['and'])
    OR      = pp.CaselessLiteral(SYMBOLS['or'])
    NOT     = pp.CaselessLiteral(SYMBOLS['not'])
    NA      = pp.CaselessLiteral(SYMBOLS['na'])
    NUM     = pp.pyparsing_common.number
    VAR     = (pp.CaselessLiteral(SYMBOLS['var']) +
               pp.pyparsing_common.integer).setParseAction(parseVariable)

    # a single conditional statement:
    # "variable comparison_operator value"
    NUMCOND = pp.Group(VAR + CMP + NUM).setParseAction(parseCondition)
    NACOND  = pp.Group(VAR + EQS + NA) .setParseAction(parseCondition)
    COND    = NUMCOND ^ NACOND

    # the infixNotation helper does the heavy
    # lifting for boolean operations and precedence
    parser = pp.infixNotation(
        COND,
        [(NOT, 1, pp.opAssoc.RIGHT, parseLogicalNot),
         (AND, 2, pp.opAssoc.LEFT , parseLogical),
         (OR,  2, pp.opAssoc.LEFT,  parseLogical)])

    makeParser.parser = parser
    return parser


def parseVariable(toks):
    """Called by the parser created by :func:`makeParser`. Parses a variable
    identifier, returning an integer ID.
    """
    return toks[1]


def _not(op,       *args): return ~ op(*args)              # noqa
def _and(op1, op2, *args): return op1(*args) & op2(*args)  # noqa
def _or( op1, op2, *args): return op1(*args) | op2(*args)  # noqa


def parseLogicalNot(toks):
    """Called by the parser created by :func:`makeParser`. Parses an expression
    of the form ``not expression``, where ``not`` is the corresponding symbol
    in the :attr:`OPERATIONS` dictionary, and ``expression`` is a conditional
    statement or logical expression.

    Returns a function which can be used to evaluate the expression.
    """

    operation = toks[0][0]
    operand   = toks[0][1]

    log.debug('Parsing logical: %s %s', operation, operand)

    fn           = ft.partial(_not, operand)
    fn.ftype     = 'logical_not'
    fn.operation = operation
    fn.operand   = operand

    return fn


def parseLogical(toks):
    """Called by the parser created by :func:`makeParser`. Parses an
    expression of the form ``expression1 [and|or] expression2``, where
    ``and``/``or`` are the corresponding symbols in the :attr:`OPERATIONS`
    dictionary, and ``expression1`` and ``expression2`` are conditional
    statements or logical expression.

    Returns a function which can be used to evaluate the expression.
    """

    operand1  = toks[0][0]
    operation = toks[0][1]
    operand2  = toks[0][2]

    log.debug('Parsing logical %s %s %s', operand1, operation, operand2)

    if   operation == SYMBOLS['and']: fn = _and
    elif operation == SYMBOLS['or']:  fn = _or

    fn           = ft.partial(fn, operand1, operand2)
    fn.ftype     = 'logical'
    fn.operation = operation
    fn.operand1  = operand1
    fn.operand2  = operand2

    return fn


def _isna( var, val, dt, data): return dt[:, data[var]].isna()  # noqa
def _notna(var, val, dt, data): return dt[:, data[var]].notna() # noqa
def _eq(   var, val, dt, data): return dt[:, data[var]] == val  # noqa
def _ne(   var, val, dt, data): return dt[:, data[var]] != val  # noqa
def _gt(   var, val, dt, data): return dt[:, data[var]] >  val  # noqa
def _ge(   var, val, dt, data): return dt[:, data[var]] >= val  # noqa
def _lt(   var, val, dt, data): return dt[:, data[var]] <  val  # noqa
def _le(   var, val, dt, data): return dt[:, data[var]] <= val  # noqa


def parseCondition(toks):
    """Parses a conditional statement of the form::

        variable operation value

    where:
      - ``variable`` is a variable identifier
      - ``operation`` is a comparison operation
      - ``value`` is a numeric value

    Returns a function which can be used to evaluate the conditional statement.
    """
    toks      = toks[0]
    variable  = toks[0]
    operation = toks[1]
    value     = toks[2]

    log.debug('Parsing condition: v%s %s %s', variable, operation, value)

    if   operation == SYMBOLS['eq'] and value == 'na': fn = _isna
    elif operation == SYMBOLS['ne'] and value == 'na': fn = _notna
    elif operation == SYMBOLS['eq']:                   fn = _eq
    elif operation == SYMBOLS['ne']:                   fn = _ne
    elif operation == SYMBOLS['ge']:                   fn = _ge
    elif operation == SYMBOLS['gt']:                   fn = _gt
    elif operation == SYMBOLS['le']:                   fn = _le
    elif operation == SYMBOLS['lt']:                   fn = _lt

    fn           = ft.partial(fn, variable, value)
    fn.ftype     = 'condition'
    fn.operation = operation
    fn.variable  = variable
    fn.value     = value

    return fn



#!/usr/bin/env python
#
# test_processing.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import itertools as it
import multiprocessing as mp
import textwrap as tw
from unittest import mock
from collections import defaultdict

import numpy  as np
import pandas as pd
import           pytest

import funpack.processing as processing
import funpack.importing  as importing
import funpack.loadtables as loadtables
import funpack.custom     as custom
import funpack.util       as util


from . import (clear_plugins,
               tempdir,
               gen_test_data,
               gen_DataTableFromDataFrame,
               gen_tables)


def  test_processData():           _test_processData(False)
def  test_processData_lowMemory(): _test_processData(True)
def _test_processData(lowMemory):

    custom.registerBuiltIns()

    sparse      = np.random.random(100)
    sparse[:50] = np.nan
    series1     = np.sin(np.linspace(0, np.pi * 6, 100))
    series2     = series1 + np.random.random(100)
    corr        = pd.Series(series1).corr(pd.Series(series2))

    vartable, proctable, cattable = gen_tables(range(1, 4))[:3]

    procs = [
        'removeIfSparse(60)',
        'removeIfRedundant({:0.2f})'.format(corr - 0.01)]

    procs  = [processing.parseProcesses(p, 'processor') for p in procs]
    procs  = [{p[0].name : p[0]} for p in procs]

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()

        df = pd.DataFrame({'1-0.0' : sparse,
                           '2-0.0' : series1,
                           '3-0.0' : series2},
                          index=np.arange(1, 101))
        df.index.name = 'eid'
        df.to_csv('data.txt')

        dtable, _ = importing.importData('data.txt',
                                         vartable,
                                         proctable,
                                         cattable,
                                         lowMemory=lowMemory,
                                         pool=pool,
                                         mgr=mgr)

        dtable.proctable['Variable'] = [('all_independent', []),
                                        ('all', [])]
        dtable.proctable['Process']  = procs
        processing.processData(dtable)
        assert [c.name for c in dtable.allColumns[1:]] == ['2-0.0']

        dtable, _ = importing.importData('data.txt',
                                         vartable,
                                         proctable,
                                         cattable,
                                         lowMemory=lowMemory,
                                         pool=pool,
                                         mgr=mgr)
        dtable.proctable['Variable'] = [('vids', [2, 3]),
                                        ('vids', [1, 2])]
        dtable.proctable['Process']  = procs
        processing.processData(dtable)
        assert [c.name for c in dtable.allColumns[1:]] == \
            ['1-0.0', '2-0.0', '3-0.0']
        df = None
        mgr = None
        pool = None
        dtable = None


def test_parseProcesses_parser():

    parser = processing.makeParser()

    with mock.patch('funpack.custom.exists', return_value=True):

        p = processing.parseProcesses('blah', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == ()
        assert p.kwargs == {}
        print(p)

        p = processing.parseProcesses('blah()', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == ()
        assert p.kwargs == {}

        p = processing.parseProcesses('blah(1, 2)', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == (1, 2)
        assert p.kwargs == {}

        p = processing.parseProcesses('blah("a", \'b\', 3)', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == ('a', 'b', 3)
        assert p.kwargs == {}

        p = processing.parseProcesses('blah(a=1, b=2, c=3)', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == ()
        assert p.kwargs == {'a' : 1, 'b' : 2, 'c' : 3}

        p = processing.parseProcesses('blah("a", b=2, c=3)', 'none')
        assert len(p)   == 1
        p = p[0]
        assert p.name   == 'blah'
        assert p.args   == ('a',)
        assert p.kwargs == {'b' : 2, 'c' : 3}


@clear_plugins
def test_parseProcesses_run():

    called = {}

    @custom.processor('boo')
    def boo(a, b):
        assert a == 1 and b == 'hah'
        called['boo'] = True

    @custom.processor('foo')
    def foo(c, d):
        assert c == 'wah' and d == 4
        called['foo'] = True

    @custom.cleaner('moo')
    def moo(*a):
        called['moo'] = True

    @custom.cleaner('woo')
    def woo(x, y):
        assert x == 10 and y == 20
        called['woo'] = True

    @custom.processor('hoo')
    def hoo(truearg, falsearg, nonearg):
        assert truearg is True
        assert falsearg is False
        assert nonearg is None
        called['hoo'] = True

    procs  = processing.parseProcesses(
        'boo(1, \'hah\'), foo("wah", 4), hoo(True, False, None)',
        'processor')
    pprocs = processing.parseProcesses(
        'moo, moo(), woo(10, 20)',
        'cleaner')

    assert procs[0].name == 'boo'
    assert procs[0].args == (1, 'hah')
    assert procs[1].name == 'foo'
    assert procs[1].args == ('wah', 4)
    assert procs[2].name == 'hoo'
    assert procs[2].args == (True, False, None)

    procs[0].run()
    procs[1].run()
    procs[2].run()

    assert called['boo']
    assert called['foo']
    assert called['hoo']

    assert pprocs[0].name == 'moo'
    assert pprocs[1].name == 'moo'
    assert pprocs[2].name == 'woo'
    assert pprocs[0].args == ()
    assert pprocs[1].args == ()
    assert pprocs[2].args == (10, 20)

    pprocs[0].run()
    assert called['moo']
    called.clear()
    pprocs[1].run()
    assert called['moo']
    pprocs[2].run()
    assert called['woo']

    with pytest.raises(processing.NoSuchProcessError):
        processing.parseProcesses('gurh', 'processor')
    with pytest.raises(processing.NoSuchProcessError):
        processing.parseProcesses('boo', 'cleaner')
    with pytest.raises(processing.NoSuchProcessError):
        processing.parseProcesses('moo', 'processor')


@clear_plugins
def test_processData_variable_types():

    procfile = tw.dedent("""
    Variable\tProcess
    1:3\trun_vids
    all\trun_all
    all_independent\trun_all_independent
    all_except,1,2,3\trun_all_except
    all_independent_except,4,5,6\trun_all_independent_except
    """).strip()

    called_on = defaultdict(list)

    @custom.processor()
    def run_vids(dtable, vids):
        called_on['vids'].append(vids)

    @custom.processor()
    def run_all(dtable, vids):
        called_on['all'].append(vids)

    @custom.processor()
    def run_all_independent(dtable, vids):
        called_on['all_independent'].append(vids)

    @custom.processor()
    def run_all_except(dtable, vids):
        called_on['all_except'].append(vids)

    @custom.processor()
    def run_all_independent_except(dtable, vids):
        called_on['all_independent_except'].append(vids)

    with tempdir():
        open('processing.tsv', 'wt').write(procfile)
        gen_test_data(6, 50, 'data.tsv')

        proctable = loadtables.loadProcessingTable('processing.tsv')
        vartable, _, cattable = gen_tables(range(1, 7))[:3]
        dtable, _ = importing.importData('data.tsv',
                                         vartable,
                                         proctable,
                                         cattable)
        processing.processData(dtable)

        assert called_on['vids']                   == [[1, 2, 3]]
        assert called_on['all']                    == [[1, 2, 3, 4, 5, 6]]
        assert called_on['all_independent']        == [[1], [2], [3],
                                                       [4], [5], [6]]
        assert called_on['all_except']             == [[4, 5, 6]]
        assert called_on['all_independent_except'] == [[1], [2], [3]]


@clear_plugins
def test_processData_returnValues():

    @custom.processor()
    def nothing(dtable, vids):
        return None

    @custom.processor()
    def remove(dtable, vids):
        cols = list(it.chain(*[dtable.columns(v) for v in vids]))
        return cols

    @custom.processor()
    def add(dtable, vids):
        newseries = []
        newvids   = []
        for v in vids:
            col = dtable.columns(v)[0]
            data = dtable[:, col.name]
            newseries.append(pd.Series(data + 10,
                                       name='{}-0.0'.format(v * 10)))
            newvids.append(v * 10)
        return newseries, newvids

    @custom.processor()
    def add_and_remove(dtable, vids):
        remcols   = []
        newseries = []
        newvids   = []
        for v in vids:
            col = dtable.columns(v)[0]
            data = dtable[:, col.name]
            newseries.append(pd.Series(data + 10,
                                       name='{}-0.0'.format(v * 10)))
            newvids.append(v * 10)
            remcols.append(col)
        return remcols, newseries, newvids

    @custom.processor()
    def add_and_remove_meta(dtable, vids):
        remcols   = []
        newseries = []
        newvids   = []
        newmeta   = []
        for v in vids:
            col = dtable.columns(v)[0]
            data = dtable[:, col.name]
            newseries.append(pd.Series(data + 10,
                                       name='{}-0.0'.format(v * 10)))
            newvids.append(v * 10)
            remcols.append(col)
            newmeta.append(v * 10)
        return remcols, newseries, newvids, newmeta

    procfile = tw.dedent("""
    Variable\tProcess
    1:3\tremove
    4:6\tadd
    7:9\tadd_and_remove
    10:12\tnothing
    13:15\tadd_and_remove_meta
    """).strip()

    with tempdir():
        open('processing.tsv', 'wt').write(procfile)
        gen_test_data(15, 50, 'data.tsv')

        proctable = loadtables.loadProcessingTable('processing.tsv')
        vartable, _, cattable = gen_tables(range(1, 16))[:3]
        dtable, _ = importing.importData('data.tsv',
                                         vartable,
                                         proctable,
                                         cattable)

        processing.processData(dtable)

        gotcols = [c.name for c in dtable.allColumns[1:]]
        expcols = ['{}-0.0'.format(v)
                   for v in [4, 5, 6,
                             10, 11, 12,
                             40, 50, 60,
                             70, 80, 90,
                             130, 140, 150]]

        assert sorted(expcols) == sorted(gotcols)


@clear_plugins
def test_processData_metadata():

    @custom.metaproc()
    def modmeta(dtable, vid, val):
        return val * 10

    @custom.processor()
    def process(dtable, vids):
        newseries = []
        newvids   = []
        newmeta   = []
        for v in vids:
            col = dtable.columns(v)[0]
            data = dtable[:, col.name]
            newseries.append(pd.Series(data + 10,
                                       name='{}-0.0'.format(v * 10)))
            newvids.append(v * 10)
            newmeta.append(v * 10)
        return [], newseries, newvids, newmeta

    procfile = tw.dedent("""
    Variable\tProcess
    1\tprocess
    2\tprocess(metaproc='modmeta')
    """).strip()

    with tempdir():
        open('processing.tsv', 'wt').write(procfile)
        gen_test_data(2, 50, 'data.tsv')

        proctable = loadtables.loadProcessingTable('processing.tsv')
        vartable, _, cattable = gen_tables([1, 2])[:3]
        dtable, _ = importing.importData('data.tsv',
                                         vartable,
                                         proctable,
                                         cattable)
        processing.processData(dtable)
        gotcols = [c.name for c in dtable.allColumns[1:]]
        expcols = ['{}-0.0'.format(v) for v in [1, 2, 10, 20]]

        assert sorted(expcols) == sorted(gotcols)

        col1, col2, col10, col20 = dtable.allColumns[1:]

        assert col1 .metadata is None
        assert col2 .metadata is None
        assert col10.metadata == 10
        assert col20.metadata == 200

def test_processData_existing_column():

    @custom.processor()
    def myproc(dtable, vids):
        remcols   = []
        newseries = []
        newvids   = []

        vid  = vids[0]
        col  = dtable.columns(vid)[0]
        data = dtable[:, col.name]

        # add a new column with the
        # same name as the existing
        # one, and remove the existing
        # one.
        newseries.append(pd.Series(data + 10, name=col.name))
        newvids.append(vid)
        remcols.append(col)

        return remcols, newseries, newvids

    data = np.random.randint(0, 10, 50)
    df    = pd.DataFrame({'eid'   : np.arange(1, 51),
                          '1-0.0' : data}).set_index('eid')
    dt    = gen_DataTableFromDataFrame(df)

    procs = processing.parseProcesses('myproc', 'processor')
    procs = {procs[0].name : procs[0]}

    dt.proctable.loc[1, ['Variable', 'Process']] = [('vids', [1]), procs]

    # make sure the variable metadata
    # is preserved for the new column
    dt.vartable.at[1, 'Type']        = util.CTYPES.categorical_single
    dt.vartable.at[1, 'Description'] = 'Super cool'

    processing.processData(dt)

    assert np.all(dt.vartable.index)       == [1]
    assert [c.name for c in dt.allColumns] == ['eid', '1-0.0']
    assert dt.vartable.at[1, 'Type'] == util.CTYPES.categorical_single
    assert dt.vartable.at[1, 'Description'] == 'Super cool'



#!/usr/bin/env python

import pandas as pd

from funpack import sniffer, loader, Column

# Sniffer and loader functions are defined in
# pairs. A pair is denoted by its functions
# being given the same label, passed to the
# @sniffer/@loader decorators
# ("my_datefile_loader" in this example). The
# function names are irrelevant.

@sniffer('my_datefile_loader')
def columns_datefile(infile):

    # A sniifer function must return a
    # sequence of Column objects which
    # describe the file (after it has
    # been loaded by the loader function).
    #
    # The values passed to a Column object
    # are:
    #   - file name
    #   - column name
    #   - column index (starting from 0)
    return [Column(infile, 'eid',              0),
            Column(infile, 'acquisition_date', 1)]

@loader('my_datefile_loader')
def load_datefile(infile):

    def create_date(row):
        return pd.Timestamp(row['year'], row['month'], row['day'])

    df = pd.read_csv(infile, index_col=0, delim_whitespace=True)

    df['acquisition_date'] = df.apply(create_date, axis=1)

    df.drop(['year', 'month', 'day'], axis=1, inplace=True)

    return df
